/*
 * Copyright (c) 1990,1991 Regents of The University of Michigan.
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appears in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation, and that the name of The University
 * of Michigan not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission. This software is supplied as is without expressed or
 * implied warranties of any kind.
 *
 *	Research Systems Unix Group
 *	The University of Michigan
 *	c/o Mike Clark
 *	535 W. William Street
 *	Ann Arbor, Michigan
 *	+1-313-763-0525
 *	netatalk@itd.umich.edu
 */
/*
 * Contributed by Thomas Tornblom
 * thomas@nexus.com.se
 */

#ifdef USEPPD

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <string.h>
#include <ctype.h>

#include <netatalk/at.h>
#include <atalk/atp.h>
#include "pap.h"
#include "papd.h"
#include "lw.h"

char			*malloc();
char			*realloc();
char			*index();
char			*rindex();
char			*copyarg();
struct ppd_feature	*ppd;

extern int		debugflg;

read_ppd_file(ppdfilename, ppflag)
    char	*ppdfilename;
    int		ppflag;
{
    char		line[100], *p;
    struct ppd_feature	*pp;
    FILE		*f;

    if (( f = fopen( ppdfilename, "r" )) == NULL ) {
	return -1;
    }
    
    while (fgets(line, sizeof(line) -1, f)) {

	if (((line[0] == '*' && isalpha(line[1])) || ppflag) &&
	    (p = index(line, ':'))) {
	    *p++ = 0;

	    if (strncmp(line, "*Font ", sizeof("*Font ") - 1) == 0) {
		if (addfont( &line[sizeof( "*Font ")-1]) < 0)
		    return -1;
	    } else {
/*
 * Should we really skip the space? It seems natural to skip it but
 * the Adobe paper says "It is important that the spooler simply returns
 * the value as is after the colon of the feature/option and not do any
 * additional processing." The LW 8.0 can't pick some of the values if
 * they start with a space. I've been pragmatic here and ignored the
 * this statement in the paper, skip the space to to keep the mac happy ;-)
 */
		while (*p == ' ')
		    p++;

		if ((pp = (struct ppd_feature *)
		     malloc(sizeof(struct ppd_feature))) == 0) {
		    return -1;
		}
		pp->ppdf_feature = copyarg(line);
		pp->ppdf_value = copyarg(p);
		pp->ppdf_next = ppd;
		ppd = pp;
	    }
	}
    }
    fclose(f);
#ifdef EBUG
    {
	for ( pp = ppd; pp != NULL; pp = pp->ppdf_next ){
	    fprintf(stderr, "Found [%s], [%s]\n", pp->ppdf_feature,
		    pp->ppdf_value);
	}
    }
#endif
    return 0;
}


free_featurelist()
{
    struct ppd_feature	*nextpp;

    while ( ppd != NULL ) {
	nextpp = ppd->ppdf_next;
	free( ppd->ppdf_feature );
	free( ppd->ppdf_value );
	free( ppd );
	ppd = nextpp;
    }
}


q_feature( ph, qt, bufp, blen )
    PAP			ph;
    struct query_table	*qt;
    char		**bufp;
    int			*blen;
{
    char		*s;
    struct ppd_feature	*pp;

#ifdef EBUG
    if ( debugflg ) {
	fprintf( stderr, "entered q_feature, query <%s>\n",
		qt->query_params);
	fflush( stderr );
    }
#endif EBUG

    if (( s = swallow_query( bufp, blen )) == NULL
      || ( s = index( s, ' ' )) == NULL ) {
#ifdef EBUG
	if ( debugflg ) {
	    fprintf( stderr, "q_feature returning -1\n" );
	    fflush( stderr );
	}
#endif EBUG
	return -1;
    }
    ++s;
    /* send back reply */
    pp = ppd;

    for (pp = ppd;
	 pp && strncmp(pp->ppdf_feature, qt->query_params,
		       strlen(pp->ppdf_feature)) != 0;
	 pp = pp->ppdf_next);

    if (pp)
	s = pp->ppdf_value;

#ifdef EBUG
    if ( debugflg ) {
	fprintf( stderr, "q_feature is <%.*s> -- sending it\n", linelen(s)+1, s );
	fflush( stderr );
    }
#endif EBUG

    if ( pap_write( ph, s, linelen( s )+1, 0 ) < 0 ) {
	return -1;
    }
    return 0;
}
#endif USEPPD
