/*
 * Copyright (c) 1990,1991 Regents of The University of Michigan.
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appears in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation, and that the name of The University
 * of Michigan not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission. This software is supplied as is without expressed or
 * implied warranties of any kind.
 *
 *	Research Systems Unix Group
 *	The University of Michigan
 *	c/o Mike Clark
 *	535 W. William Street
 *	Ann Arbor, Michigan
 *	+1-313-763-0525
 *	netatalk@itd.umich.edu
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <string.h>
#include <ctype.h>

#include <netatalk/at.h>
#include <atalk/atp.h>
#include "pap.h"
#include "papd.h"
#include "lw.h"

char			*malloc();
char			*realloc();
char			*index();
char			*rindex();
char			*copyarg();

extern int		debugflg;
extern struct font	*fontlist;
extern long		dsc_vers;


addfont( fontname )
char	*fontname;
{
    struct font	*f;

#ifdef EBUG
    if ( debugflg ) {
	fprintf( stderr, "addfont: <%s>\n", fontname );
	fflush( stderr );
    }
#endif
    
    if ((f = (struct font *) malloc(sizeof(struct font))) == NULL) {
	return -1;
    }

    f->font_next = fontlist;

    if ((f->font_name = copyarg( fontname )) == NULL) {
	free( f );
	return -1;
    }
    fontlist = f;
    return 0;
}


readfontfile( fontfilename )
    char	*fontfilename;
{
    char	fontname[ 65 ];
    char	*p;
    FILE	*f;

    if (( f = fopen( fontfilename, "r" )) == NULL ) {
	return( -1 );
    }

    while ( fgets( fontname, 64, f ) != NULL ) {
	if ( p = index(fontname, '\n')) {
	    *p = '\0';			/* get rid of newline */
	}

	if ( addfont( fontname ) == -1) {
	    return( -1 );
	}
    }

    fclose( f );
    return( 0 );
}


free_fontlist()
{
    struct font	*nextfp;

    while ( fontlist != NULL ) {
	nextfp = fontlist->font_next;
	free( fontlist->font_name );
	free( fontlist);
	fontlist = nextfp;
    }
}


/*
 * The normal thing is to return the reply in the reverse order, you
 * know, PostScript being stack oriented and everything, therefore
 * we call this function recursively. We also pack up and send the
 * reply in just one response packet
 */
char *
q_font1( query, reply )
    char		*query, *reply;
{
/* handle a font query */ 
    char	*p;
    struct font	*f;

    if ( *query != '\0' ) {
	for ( p = query; *p && *p != ' '; p++);

	if ( *p )
	    reply = q_font1( p+1, reply ); /* recurse */

	for ( f = fontlist; f; f = f->font_next ) {
#ifdef EBUG
	    if ( debugflg ) {
		fprintf( stderr, "q_font checking against <%s>\n",
			f->font_name );
		fflush( stderr );
	    }
#endif EBUG
	    if ( strncmp( query, f->font_name, p - query ) == 0 )
		break;
	}

	if ( dsc_vers >= DSC_2_1 ) {

	    *reply++ = '/';

	    while ( query != p ) {	/* copy the fontname */
		*reply++ = *query++;
	    }
	    strcpy( reply, f ? ":Yes\n" : ":No\n");

	    while (*reply)
		reply++;
	} else {
	    *reply++ = f ? '1' : '0';
	    *reply++ = '\n';
	}
    }

    return reply;
}

q_font( ph, qt, bufp, blen )
    PAP			ph;
    struct query_table	*qt;
    char		**bufp;
    int			*blen;
{
/* handle a font query */ 
    static char	reply[ 256 ];
    char	*p;

    if ( swallow_query( bufp, blen ) == NULL ) {
	return( -1 );
    }

#ifdef EBUG 
    if ( debugflg ) {
	fprintf( stderr, "q_font list is <%s>\n", qt->query_params );
	fflush( stderr );
    }
#endif EBUG

    p = q_font1( qt->query_params, reply);

    if ( dsc_vers >= DSC_2_1 ) {
	*p++ = '*';
	*p++ = '\n';			/* EOL; */

#ifdef EBUG
	if ( debugflg ) {
	    fprintf( stderr, "q_font reply is <%.*s>\n",
		    p - reply, reply);
	    fflush( stderr );
	}
#endif
	if ( pap_write( ph, reply, p - reply, 0 ) < 0 ) {
	    return( -1 );
	}
    }

    return( 0 );
}


q_fontlist( ph, qt, bufp, blen )
    PAP			ph;
    struct query_table	*qt;
    char		**bufp;
    int			*blen;
{
/* handle a fontlist query */ 
    static char	reply[ 128 ];
    char	*rp;
    char	*fp;
    struct font	*f;

    if ( swallow_query( bufp, blen ) == NULL ) {
	return( -1 );
    }
    /* send back reply */
    if ( fontlist != NULL ) {
#ifdef EBUG
	if ( debugflg ) {
	    fprintf( stderr, "q_fontlist sending list of fonts\n" );
	    fflush( stderr );
	}
#endif EBUG
	for ( f = fontlist; f != NULL; f = f->font_next ) {
	    rp = reply;
	    fp = f->font_name;
	    if ( dsc_vers >= DSC_2_1 ) {
		*rp++ = '/';
	    }

	    while (*fp)
		*rp++ = *fp++;

	    *rp++ = '\n';

#ifdef EBUG
	if ( debugflg ) {
	    fprintf( stderr, "q_fontlist: <%.*s>\n",
		    rp - reply, reply);
	    fflush( stderr );
	}
#endif EBUG
	    if ( pap_write( ph, reply, rp - reply, 0 ) < 0
		    || pap_get_request( ph ) != 0 ) {
		return( -1 );
	    }
	}
    }
    reply[ 0 ] = '*';
    reply[ 1 ] = '\n'; /* EOL */
    if ( pap_write( ph, reply, 2, 0 ) < 0 ) {
	return( -1 );
    }

    return( 0 );
}
